// IP checksumming
//
// addr IN  offset of data
// len  IN  length of data in BYTES (!!!)
// returns checksum of data
inline unsigned short in_cksum(unsigned short *addr, unsigned int len) {
	unsigned int	sum;

#if !defined(__i386__)
#warning "Using C version for IP checksum"
	// Code taken from "Unix Network Programming" by W. Richard Stevens
	unsigned short	answer;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	sum=0;
	answer=0;
	while (len > 1)  {
		sum += *addr++;
		len -= 2;
	}

	/* 4mop up an odd byte, if necessary */
	if (len == 1) {
		*(unsigned char *)(&answer) = *(unsigned char *)addr ;
		sum += answer;
	}

	/* 4add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
#else
#warning "Using i386 asm version for IP checksum"
	// This code comes from the Linux kernel (see asm/checksum.h) and only
	// works for IP packets whose length is a multiple of 4, which is
	// hopefully the case here
	unsigned char *iph=(unsigned char *)addr;
	unsigned int ihl=len/4;

	__asm__ __volatile__(
"		movl (%1), %0		\n"
"		subl $4, %2			\n"
"		jbe 2f				\n"
"		addl 4(%1), %0		\n"
"		adcl 8(%1), %0		\n"
"		adcl 12(%1), %0		\n"
"1:	  adcl 16(%1), %0		\n"
"		lea 4(%1), %1		\n"
"		decl %2				\n"
"		jne 1b				\n"
"		adcl $0, %0			\n"
"		movl %0, %2			\n"
"		shrl $16, %0		\n"
"		addw %w2, %w0		\n"
"		adcl $0, %0			\n"
"		notl %0				\n"
"2:							\n"
		
	/* Since the input registers which are loaded with iph and ipl
	   are modified, we must also specify them as outputs, or gcc
	   will assume they contain their original values. */
	: "=r" (sum), "=r" (iph), "=r" (ihl)
	: "1" (iph), "2" (ihl));
	return(sum);
#endif

}

// vim:tw=78:ts=4:sw=4
