// This is a test server
//
// Author: Pierre Lombard <pierre.lombard@imag.fr> 2002/07 GPLv2

#include "../sure_sock.h"
#include "../sure_sock.c"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include "nfspd_stub.h"

#define ERR(fmt,args...) do { int tmp = errno; fprintf(stderr, fmt "\n" ,## args); errno = tmp; } while (0)

static int sock_ctl = -1;
static int new_sock = -1;

static void sigendhdl(int sig) {
	printf("\nReceived signal %d - exiting\n", sig);
	exit(1);
}

static void terminate() {
	struct sockaddr_un sun;
	socklen_t sun_size = sizeof(sun); 
	printf("\nTermination handler called... ");
	fflush(stdout);

	if (new_sock >= 0) {
		close(new_sock);
	}
	if (-1 == getsockname(sock_ctl, (struct sockaddr *) &sun, &sun_size)) {
		ERR("cannot getsockname(sock_ctl) errno=%d", errno);
	}
	if (-1 == unlink(sun.sun_path)) {
		ERR("cannot unlink(sock_ctl) errno=%d", errno);
	}
	if (-1 == close(sock_ctl)) {
		ERR("cannot close(sock_ctl) errno=%d", errno);
	}
	printf("done\n");
}

static int nfsp_iodctl (int fd) {
	int result, ret;
	struct message_t mess;

	ret = sure_recv (fd, &mess, sizeof(mess), MSG_NOSIGNAL);
	if (ret != sizeof(mess)) {
		return -1;
	}

	if (mess.magic != NFSP_MAGIC) {
		ERR("not enough magic");
		return -1;
	}

	result = 0;
	switch (mess.fnct) {
	case NFSP_ADD_IOD:
		do {
			unsigned int ipnum = ntohl(mess.addr.sin_addr.s_addr);
			unsigned char *ip = (unsigned char *) &ipnum;
			ERR("ADD_IOD %d.%d.%d.%d:%d", ip[0], ip[1], ip[2], ip[3], ntohs(mess.addr.sin_port));
		} while (0);
		break;
	case NFSP_FREE_IOD:
		ERR("FREE_IOD ih = %d", mess.ih);
		break;
	case NFSP_BIND:
		ERR("BIND iviod = %d ih = %d", mess.iviod, mess.ih);
		break;
	case NFSP_UNBIND:
		ERR("UNBIND ih = %d", mess.ih);
		break;
	case NFSP_SETNBVIODS:
		ERR("SETNBVIODS h = %d", mess.nbviods);
		break;
	case NFSP_GETNBVIODS:
		ERR("GETNBVIODS");
		ret = random();
		break;
	case NFSP_DROP:
		ERR("DROP");
		break;
	case NFSP_FORWARD:
		ERR("FORWARD");
		break;
	case NFSP_FLUSH:
		ERR("FLUSH");
		break;
	default:
		ERR("UNKNOWN %d", mess.fnct);
		result = -1;
	}

	ret = sure_send (fd, &result, sizeof (result), MSG_NOSIGNAL);
	if (ret != sizeof (result)) {
		return -1;
	}
	return 0;
}


int main(int argc, char *argv[]) {
	int ret;
	struct sockaddr_un sun;
	char *home;

	struct sockaddr_un new_sun;
	socklen_t new_sun_len;

	printf("Registering atexit() handler... ");
	ret = atexit(terminate);
	printf("done (ret = %d)\n", ret);
	
	printf("Registering signal() handler... ");
	signal(SIGHUP, sigendhdl);
	signal(SIGINT, sigendhdl);
	signal(SIGTERM, sigendhdl);
	signal(SIGKILL, sigendhdl);
	printf("done\n");
	
	sock_ctl = socket(PF_UNIX, SOCK_STREAM, 0);
	if (sock_ctl < 0) {
		ERR("cannot open UNIX socket errno=%d", errno);
		exit(1);
	}
	sun.sun_family = AF_UNIX;
	home = getenv(SOCK_CTL_ENV_VAR);
	if (!home) {
		ERR("%s must be set", SOCK_CTL_ENV_VAR);
		exit(1);
	}
	if (sizeof(sun.sun_path) < strlen(home) + 1 + strlen(SOCK_CTL_NAME) + 1) {
		ERR("path to UNIX control socket too long (>%d)", sizeof(sun.sun_path));
		exit(1);
	}
	strcpy(sun.sun_path, home);
	strcat(sun.sun_path, "/");
	strcat(sun.sun_path, SOCK_CTL_NAME);

	if (-1 == bind(sock_ctl, (struct sockaddr *) &sun, sizeof(sun))) {
		ERR("cannot bind() UNIX socket errno=%d", errno);
		exit(1);
	}

	if (-1 == listen(sock_ctl, 5)) {
		ERR("cannot listen() UNIX socket errno=%d", errno);
		exit(1);
	}


	while(1) {
	       	new_sun_len = sizeof(new_sun);
		new_sock = accept(sock_ctl, (struct sockaddr *)&new_sun, &new_sun_len);
		if (new_sock == -1) {
			if (errno == EINTR) {
				printf("\naccept interrupted by signal\n");
			}
			continue;
		}

		do {
		} while (0 <= nfsp_iodctl(new_sock));
		close(new_sock);
	}

}

