// Author: Pierre Lombard <pierre.lombard@imag.fr>
// Date: 200211xx
// License: GPLv2

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>

#include "recursdir.h"
#include "xalloc.h"
#include "access_config.h"
#include "timed_connect.h"

#include "netproto.h"
#include "xalloc.h"
#include "server_common.h"
#include "sure_sock.h"

#define DECLARE_get_type(the_type)                                      \
static inline int get_##the_type(int sock, the_type *valp) {            \
        int ret;                                                        \
        ret = sure_recv(sock, valp, sizeof(*valp), MSG_NOSIGNAL);       \
        return ret != sizeof(*valp);                                    \
}

DECLARE_get_type(int16_t)
DECLARE_get_type(int32_t)
DECLARE_get_type(uint16_t)
DECLARE_get_type(uint32_t)


#define DECLARE_put_type(the_type)                                      \
static inline int put_##the_type(int sock, the_type *valp) {            \
        int ret;                                                        \
        ret = sure_send(sock, valp, sizeof(*valp), MSG_NOSIGNAL);       \
        return ret != sizeof(*valp);                                    \
}

DECLARE_put_type(int16_t)
DECLARE_put_type(int32_t)
DECLARE_put_type(uint16_t)
DECLARE_put_type(uint32_t)


int
main(int argc, char *argv[])
{
	char *metafiles_path;
	struct valid_file_t *files_array;
	int files_array_size;
	struct viod_t *viods;
	int viods_size;
	int j, i;

	metafiles_path = argv[1];

	if (get_config(&viods, &viods_size)) {
		fprintf(stderr, "FATAL ERROR: cannot get system configuration - exiting\n");
		exit(1);
	}

	mk_valid_inode_array(metafiles_path, &files_array, &files_array_size);

	printf("metafiles_path=%s files_array=%p files_array_size=%d",
			metafiles_path, files_array, files_array_size);

	// iterate on the viods
	for (j = 0; j < viods_size; j++) {
		fprintf(stdout, "==========================\n"
				"CONTACTING VIOD #%d\n"
				"--------------------------\n", j);

		for (i = 0; i < viods[j].nbiods; i++) {
			int sock;
			struct sockaddr_in sin;

			sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
			if (sock == -1) {
				fprintf(stderr, "FATAL ERROR: cannot open TCP socket\n");
				exit(1);
			}
			fprintf(stdout, "CONNECTING TO VIOD #%d/IOD #%d\n", j, i);

			memset(&sin, 0, sizeof(sin));
			sin.sin_family = viods[j].iods[i].sin.sin_family;
			sin.sin_addr = viods[j].iods[i].sin.sin_addr;
			sin.sin_port = htons(20222);

			fprintf(stdout, "iod_id=%d iod_ip=0x%08x=%s iod_port=%d\n",
					viods[j].iods[i].id,
					ntohl(sin.sin_addr.s_addr),
					inet_ntoa(sin.sin_addr),
					ntohs(sin.sin_port));

			if (timed_connect(sock, (struct sockaddr *)&sin, sizeof(sin), 10000)) {
				fprintf(stderr, "ERROR: connection failed errno=%d\n", errno);
				goto out_error;
			}
	
			{
				uint32_t magic = SCANDISK_NET_MAGIC;
				int32_t longsize;
				uint16_t size;
				uint32_t ack;
				int ret;

	       			if (put_uint32_t(sock, &magic)) {
					fprintf(stderr, "ERROR: cannot put_uint16_t() errno=%d\n", errno);
					goto out_error;
				}

				longsize = sizeof(struct valid_file_t) * files_array_size;
				if (longsize < 0 || longsize >= 0x10000) {
					fprintf(stderr, "ERROR: cannot put_uint16_t() errno=%d\n", errno);
					goto out_error;
				}
				size = longsize;
				printf("array size = %d\n", size);
	       			if (put_uint16_t(sock, &size)) {
					fprintf(stderr, "ERROR: cannot put_uint16_t() errno=%d\n", errno);
					goto out_error;
				}

				ret = sure_send(sock, files_array, size, MSG_NOSIGNAL);
				if (ret == -1) {
					fprintf(stderr, "ERROR: cannot sure_send errno=%d\n", errno);
					goto out_error;
				} else if (ret == 0) {
					fprintf(stderr, "ERROR: connection closed by peer\n");
					goto out_error;
				}

				if (get_uint32_t(sock, &ack)) {
					fprintf(stderr, "ERROR: ack not received\n");
					goto out_error;
				}

				if (ack != SCANDISK_NET_MAGIC_ACK) {
					fprintf(stderr, "ERROR: not enough ack magic 0x%08x instead of 0x%08x\n",
							ack, SCANDISK_NET_MAGIC_ACK);
					goto out_error;
				}

			}

			printf("Connection succeed!\n");

out_error:
			close(sock);
			// timed_connect

		}
		
	}

	

	// FIXME:
	//  - contact each iod
	//  - send to each the linked files list
	//  - wait the acknowledgement

	return 0;
}
